% Copyright 2018, John C. Young, All rights reserved.
%
% Construct a linear surface mesh.
% On entry, mesh constains a list of 'nodes' and 'elements'.
% The i-th row of the 'element' array contains the node indices of the i-th element. 
% Restrictions : 
%        2. Mesh must be conformal
function mesh = construct_linear_mesh(mesh)
   
   % For now, this assumes a linear mesh so that vertices and nodes are the same   
   % Also assumes all vertices are part of some face and that there are no duplicates.
   num_ele = mesh.get_num_elements;
   
   % There are many ways to do the edge construction. This is only one method.
   
   node_to_face{mesh.get_num_nodes} = []; %struct('faces',[]);
   mesh.node_to_edge{mesh.get_num_nodes} = [];  % list of edges attached to each node     
   
   %fprintf('node-to-edge %f .\n',toc(ts));
   
   %ts=tic;
   for eid = 1:num_ele
       fnodes = mesh.get_element_nodes(eid);
       for fn = fnodes'
           node_to_face{fn} = unique([node_to_face{fn},eid]);           
       end
       
   end
   mesh.node_to_face = node_to_face; 
   %fprintf('node-to-face %f .\n',toc(ts));
   
   %%% Next, we start assigning edges by looping over each elements edges. 
   %%% We have to detect if an edge has already been placed
   %ts=tic;
   num_edges = 0;
   mesh.max_edge_length = 0.0;
 
   mesh.node_to_edge{mesh.get_num_nodes} = [];
   
   for i = 1:num_ele
       
       num_ele_nodes = mesh.get_num_element_nodes(i);
       num_local_edges = num_ele_nodes;
       mesh.element{i}.face_to_edge = zeros(1,num_local_edges);       
       for n = 1:num_ele_nodes
           
           [global_edge_num,inodes] = get_global_edge_number(mesh,i,n);           
           
           if global_edge_num <= 0
               num_edges = num_edges+1;
               global_edge_num = num_edges;               
               mesh.edges{global_edge_num}.faces = [];
               n1 = min(inodes);
               n2 = max(inodes);               
               mesh.edges{global_edge_num}.node = [n1,n2];
               mesh.max_edge_length = max(mesh.max_edge_length,norm(mesh.get_node(n1)-mesh.get_node(n2)));
               
               mesh.node_to_edge{n1} = [mesh.node_to_edge{n1},global_edge_num];
               mesh.node_to_edge{n2} = [mesh.node_to_edge{n2},global_edge_num];
              
           end

           mesh.element{i}.face_to_edge(n) = global_edge_num;
           mesh.edges{global_edge_num}.faces = [mesh.edges{global_edge_num}.faces,i];
           
       end
   end
   %fprintf('global-edges %f .\n',toc(ts));      

   % Sanity checks
   for i = 1:num_edges
       if length(mesh.edges{i}.faces) > 2
           fprintf('Edge with more than two faces detected\n');
       end
   end
   
   mesh.get_node_edges = @(nid) mesh.node_to_edge{nid};
   mesh.get_node_faces = @(nid) mesh.node_to_face{nid};
   mesh.get_edge_nodes = @(eid) mesh.edges{eid}.node;
   mesh.get_edge_faces = @(eid) mesh.edges{eid}.faces;
   mesh.get_face_nodes = @(fid) mesh.element{fid}.nodes;      
   mesh.get_face_edges = @(fid) mesh.element{fid}.face_to_edge;

   
   %fprintf('Number of edges %d\n',length(mesh.edges));
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function edge_nodes = get_edge_nodes_from_face(face_nodes,enum)

    tri = [1,2,3,1];
    quad = [1,2,4,3,1];
                 
    edge_nodes = [];
    if length(face_nodes) == 3
        edge_nodes = face_nodes(tri([enum,enum+1]));
    elseif length(face_nodes) == 4
       edge_nodes = face_nodes(quad([enum,enum+1]));
    end
    
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1. Get the nodes of the edge to be placed.
% 2. Get all of the edges connected to one of the nodes. These edges already have been placed
% 3. Loop through those edges and see if one of those edges is the current edge.
% 4. If yes, this edge has already been placed.  If not, this is a new edge.
function [global_edge_num,inodes] = get_global_edge_number(mesh,face_id,local_edge_num)

    global_edge_num = -1; % returns -1 if this edge has not been placed yet
    edge_nodes = get_edge_nodes_from_face(mesh.get_element_nodes(face_id),local_edge_num);        

    n1_elist = mesh.node_to_edge{edge_nodes(1)};
    
    for e = n1_elist
       enodes = mesh.edges{e}.node;
       if ( enodes(1) == edge_nodes(1) && enodes(2) == edge_nodes(2) ) || ...
          ( enodes(2) == edge_nodes(1) && enodes(1) == edge_nodes(2) ) 
           inodes = enodes;
           global_edge_num = e;
           return;
       end
    end
    inodes = edge_nodes; % new edge


end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
