% Copyright 2018, John C. Young, All rights reserved.
%
% Construct a simple octree for use in the coincident node finder
% nodes - a list of node (x,y,z) values to place in the octree.
% group_dim - the dimension of the groups at the finest octree level
% 
function octree = create_octree(nodes,group_dim,zerotol)

   octree.nodes = nodes;

   % Bounding box 
   octree.rmax = max(nodes,[],2);
   octree.rmin = min(nodes,[],2);
   octree.center = 0.5*(octree.rmax+octree.rmin);
   octree.bounding_box = octree.rmax-octree.rmin; 
   octree.group_dim = group_dim;
   
   % Determine a good octree depth
   % Need (bboxDim/2^num_levels) <= groupDim
   bboxDim = max([octree.bounding_box;0])*1.001; % Make a little larger to avoid floating point issues
   octree.num_levels = ceil(log(bboxDim/group_dim)/log(2.0))+1;
   octree.twotolvl = 2^(octree.num_levels-1);

   % Re-center
   halfLength = 0.5*group_dim*(2^(octree.num_levels-1));
   octree.rmin = octree.center-halfLength*ones(3,1);
   octree.rmax = octree.center+halfLength*ones(3,1);   
   octree.bounding_box = octree.rmax-octree.rmin;
   
   % Get the IJK coordinates of a point r
   octree.get_ijk = @(r) floor((r-octree.rmin)/octree.group_dim);
   
   % Compute the group id for an IJK coordinate
   octree.group_id = @(ijk) ijk(1)+octree.twotolvl*(ijk(2)+octree.twotolvl*ijk(3));
   
   % Generate the groups (Is there a better way to do this?).
   % First, generate a map of the unique group ids to an index value into the group array
   octree.group_map = containers.Map('KeyType','double','ValueType','double');
   for i = 1:size(nodes,2)
       gid = octree.group_id(octree.get_ijk(nodes(:,i)));
       if ~isKey(octree.group_map,gid)
          octree.group_map(gid) = length(octree.group_map)+1;
       end
   end
   octree.map_value = @(r) octree.group_map(octree.group_id(octree.get_ijk(r)));  
   
   % Finally, put the node ids into their groups.
   % TODO - We could check for identical nodes in the nodes array and remove them.
   octree.groups{length(octree.group_map)} = [];
   for i = 1:size(nodes,2)
       octree.groups{octree.map_value(nodes(:,i))} = [octree.groups{octree.map_value(nodes(:,i))},i];
   end
   
   % returns 0 if not in tree, else returns the id of the node.
   octree.inTree = @(r) inTree_(r,octree,zerotol);
   
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Query if the point is already in the tree
function id = inTree_(r,octree,zerotol)

   id = 0;
   gid = octree.group_id(octree.get_ijk(r));   
   if isKey(octree.group_map,gid) % If gid not in map, then not in tree
       iid = octree.group_map(gid); % else loop through group array and check point-by-point.
       for i = octree.groups{iid}
           if ( norm(octree.nodes(:,i)-r) < zerotol )
               id = i;
               break;
           end
       end
   end

end

