% Copyright 2018, John C. Young, All rights reserved.
%
% Restrictions:
%   1. Quadrilateral and triangle elements only supported.
%   2. Degenerate elements not supported.
%   3. Two elements can share at most two nodes (one edge may not be sufficient criteria).
%   4. The refined mesh but be a proper refinement of the coarse mesh.
%   5. Both meshes must be conformal and not mal-formed in any way.

function higher_order_mesh(cfilename,ffilename,ofilename,mesh_order)
     
   ts=tic;
   cmesh=read_mesh(cfilename);
   fmesh=read_mesh(ffilename);
   fprintf('Mesh read took %f s.\n',toc(ts));

   % Construct required topological information from raw meshes.
   % This includes global edges and maps of face-to-edges, edges-to-face,
   % node-to-face, and node-to-edge.  
   ts=tic;
   cmesh = construct_linear_mesh(cmesh);
   fmesh = construct_linear_mesh(fmesh);
   fprintf('Mesh construction took %f s.\n',toc(ts));
   
   % Generate a helper function that takes a list of nodes.  A given node
   % can then be queried whether or not it matches a node in the original list.
   % The original list will be the fine mesh nodes and the queried node will 
   % be a coarse mesh node.
   ts=tic;
   zerotol=1E-13;
   octree = create_octree(fmesh.nodes,fmesh.max_edge_length,zerotol);
   fprintf('Octree setup took %f s.\n',toc(ts));
   
   % Begin generating the higher-order mesh file
   oid = fopen(ofilename, 'w');
   
   % Write out the higher-order mesh nodes.  These will just be the fine mesh nodes.
   fprintf(oid,'NODES %d\n',fmesh.get_num_nodes);
   for i = 1:fmesh.get_num_nodes
       fprintf(oid,'%10d %25.15e %25.15e %25.15e\n',i,fmesh.get_node(i));
   end
   
   % Generate the higher-order mesh elements.  The higher order mesh
   % will have the same elements as the coarse mesh but with a higher-order
   % nodes representation.  This could be parallelized since each coarse mesh
   % element can be processed independently. Matlab parpool doesn't seem
   % to do a good job of parallelization here though.
   ts=tic;

   fprintf(oid,'\nELEMENTS %d\n',cmesh.get_num_elements);           
   for i = 1:cmesh.get_num_elements 
       node_list = get_higher_order_nodes_by_edges(i,cmesh,fmesh,octree,mesh_order);       
       fprintf(oid,' %10d %4d %4d\n',i,cmesh.get_element_type(i),mesh_order);
       for k=1:length(node_list)
          fprintf(oid,' %d ',node_list(k));
       end
       fprintf(oid,'\n');
   end
   
   fprintf('Higher-order mesh construction took %f s.\n',toc(ts));

%    % DEBUG
%    fprintf(oid,'\nGROUPS %d\n',1);      
%    fprintf(oid,'Surface1 %d\n',cmesh.get_num_elements);         
%    for i = 1:cmesh.get_num_elements
%        fprintf(oid,'%d ',i);
%    end
%    fprintf(oid,'\n');
   % Done
   fclose(oid);
   
   
   
end
