% Copyright 2018, John C. Young, All rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function start_data = start_topos_surf(fmesh,element_type,node_s,node_x,node_y,faces,mesh_order)

   % Each face in faces shares the start node.  However, only one of these 
   % faces will be in the constructed higher-order cell.  The proper face will
   % be found by looping over adjacent faces along x-direction until either the x-end node
   % is found or the mesh_order number of cells is processed.  
   % If the x-end node is found in exactly mesh_order faces, then we have to 
   % repeat along the y-direction.
   
   edge_order = [0,0];
   
   %modindex = [1,2,3,1,2]; % simplifies modulo-3 indexing
  
   found = @(edge_order) edge_order(1) > 0 && edge_order(2) > 0;
   
   TRIANGLE = 2;
   QUADRILATERAL = 3;   
   
      
   
   for face = faces
       if fmesh.get_element_type(face) ~= element_type; continue; end % (mixed-mesh support)
       
       % find the two local edges of the face that share the start node
       fedges = fmesh.get_face_edges(face);
       local_edge = [];
       for edge = fedges
          if edge_shares_node(edge,node_s,fmesh)
              local_edge = [local_edge,edge];
          end
       end
       if length(local_edge) ~= 2
           error('Could not determine start edge set.\n');
       end

       edge_order = [0,0];       

       % For a first-order mesh, do a quick-and-dirty set without any error checking
       if 1 == mesh_order
          if edge_shares_node(local_edge(1),node_x,fmesh) && edge_shares_node(local_edge(2),node_y,fmesh)
              edge_order = [local_edge(1),local_edge(2)];
              edge_x = local_edge(1);
              edge_y = local_edge(2);              
          elseif edge_shares_node(local_edge(2),node_x,fmesh) && edge_shares_node(local_edge(1),node_y,fmesh)
              edge_order = [local_edge(2),local_edge(1)];
              edge_x = local_edge(2);
              edge_y = local_edge(1) ;             
              
          else
             continue;
          end
          break;
       end
       
       % If here, than a higher-order refinement.
       
       edge_x = [];
       edge_y = [];

       for i = 1:2 % loop over x-direction and then y-direction. Both checks have to pass.

           edgelist = zeros(1,mesh_order);                  

           edge0 = local_edge(i);
           edge1 = local_edge(mod(i,2)+1);
           edgelist(1) = edge0;
           cur_face = face;
               
           for k = 1:mesh_order-1
               if element_type == TRIANGLE
                  [cur_face,edge0,edge1] = next_face_tri(cur_face,edge0,edge1,fmesh); % Get next face / edges                                  
               elseif element_type == QUADRILATERAL
                  [cur_face,edge0,edge1] = next_face_quad(cur_face,edge0,edge1,fmesh); % Get next face / edges                                                 
               end

               edgelist(k+1) = edge0;

               if fmesh.get_element_type(cur_face) ~= element_type; break; end % (mixed-mesh support)
           end

           if (face_shares_node(cur_face,node_x,fmesh))
               edge_order(2) = local_edge(i);
               edge_y = edgelist;
           elseif (face_shares_node(cur_face,node_y,fmesh))               
               edge_order(1) = local_edge(i);              
               edge_x = edgelist;               
           else
               break; % process next face
           end
       end

       if found(edge_order)
           break;
       end

   end


   if  ~found(edge_order)
       error('Starting topos not found.\n');
   end
   
   start_data.start_face = face;
   start_data.edge_x = edge_x;
   start_data.edge_y = edge_y;
   start_data.element_type = element_type;      
      
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% For triangles, can have either
%                                         .________. 
%                                         | \      |
%                                      e0 |   \    |
%                                         |face \  |
%       .________.________.               .__e1___\.
%       | \      | \      |               | \      |
%    e1 |   \    |e1 \    |      OR    e0 |   \    |
%       |face \  |face \  |               |face \  |
%       ._______\._______\.               ._______\.
%           e0      e0                       e1     
%
%
function [face,edge0,edge1] = next_face_tri(face,edge0,edge1,mesh)

   [face,edge0,edge1] = next_tri(face,edge0,edge1,mesh); % Get next face / edges  
   TRIANGLE = 2;
   if mesh.get_element_type(face) ~= TRIANGLE
       return; 
   end
   [face,edge1,edge0] = next_tri(face,edge1,edge0,mesh); % Get next face / edges               
   
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [face,edge0,edge1] = next_tri(face,edge0,edge1,mesh)
   
   edge0 = get_other_tri_edge(face,edge0,edge1,mesh);
   face = other_face_shared_by_edge(face,edge0,mesh);
   
   common_node = common_edge_node(edge0,edge1,mesh); 
   edge1 = 0; % for error checking
   for e = mesh.get_face_edges(face)
       if e == edge0; continue; end; % skip already found edge
       
       if edge_shares_node (e,common_node,mesh)
           edge1 = e;
           break;
       end
   end
   % Can add error checking here.
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Here we assume that by the time we enter this routine, an edge
% has exactly two faces.  This is a valid assumption on the fine mesh
% for edges produced by the searching algorithm above.
function oface = other_face_shared_by_edge(face,edge,mesh)
   faces = mesh.get_edge_faces(edge);
   if length(faces) ~= 2
       error('Invalid face list.\n');
   end
   
   %oface = setdiff(faces,face);
   if ( faces(1) == face )
       oface = faces(2);
   elseif ( faces(2) == face )
          oface = faces(1);
   else
       error('Could not find other face.\n');
   end
   
   % Can add error checking here.
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get the other edge of a triangle
function oedge = get_other_tri_edge(face,edge1,edge2,mesh)
   %edge0 = setdiff(mesh.get_face_edges(face),[edge0,edge1]);
   
   oedge = 0;
   edges = mesh.get_face_edges(face); 
   for e = edges
       if e ~= edge1 && e ~= edge2
           oedge = e;
           return;
       end
   end
   
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% For quadrliaterals, can only have
%       .________.________.
%       |        |        |      
%    e1 |        |e1      |     
%       |  face  |   face |
%       .________.________.  
%           e0        e1 

%             OR
%       .________.
%       |        |
%    e1 |  face  |
%       |        |
%       .___e0___.
%       |        |
%    e1 |        |
%       |  face  |
%       .________.
%           e0    


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [face,edge0,edge1] = next_face_quad(face,edge0,edge1,mesh)

   edge1 = opposite_quad_edge(face,edge1,mesh); % Get edge on opposite side of face
   face = other_face_shared_by_edge(face,edge1,mesh); % Junctions should be ok here
   edge0 = next_side_edge(face,edge0,edge1,mesh);
   
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function oedge = opposite_quad_edge(face,edge,mesh)
   % For a quad, the edges are ordered around the quad.  So,
   % find the local edge number of the current edge and get the second edge after it.
   
   emap = [3,4,1,2];
   oedge = 0;
   edges = mesh.get_face_edges(face);
   for i = 1:length(edges)
       if edges(i) == edge
           oedge = edges(emap(i));
           return;
       end
   end
   % Can add error checking here.
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Here we assume that by the time we enter this routine, an edge
% has exactly two faces.  This is a valid assumption on the fine mesh
% for edges produced by the searching algorithm above.
function side_edge = next_side_edge(face,pedge,oedge,mesh)

   snode = common_edge_node(oedge,pedge,mesh);
   edges = mesh.get_face_edges(face);   
   for e = edges
       if e == oedge; continue; end;
       if edge_shares_node(e,snode,mesh)
           side_edge = e;
           return;
       end
   end
   error('Could not find next side edge.\n');
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function y = edge_shares_node (edge,node,mesh)
   %edge_shares_node = @(edge,node) ~isempty(intersect(fmesh.get_edge_nodes(edge),node));
   %y = ~isempty(intersect(mesh.get_edge_nodes(edge),node));   
    enodes = mesh.get_edge_nodes(edge);
    y = (enodes(1) == node) || (enodes(2) == node);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
function y = face_shares_node(face,node,mesh)
   %face_shares_node = @(face,node) ~isempty(intersect(fmesh.get_element_nodes(face),node));
   fnodes = mesh.get_element_nodes(face);
   y = sum(fnodes == node) > 0;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%   
% Find the node shared between two edges (that are different)
function cnode = common_edge_node(edge0,edge1,mesh)
   %common_node = intersect(mesh.get_edge_nodes(edge0),mesh.get_edge_nodes(edge1));
   
   n0 = mesh.get_edge_nodes(edge0);
   n1 = mesh.get_edge_nodes(edge1);   
   
   if n0(1) == n1(1) || n0(1) == n1(2)
      cnode = n0(1);
   elseif n0(2) == n1(1) || n0(2) == n1(2)
      cnode = n0(2);
   else
      error('Could not find common node.\n');
   end
end

